#include "drvCommon.h"

#include "FileDetectAlgorithm.h"
#include "FileSearcher.h"
#include "StringUtils.h"
#include "ServiceTableUtils.h"
#include "CustomFileApi.h"

#include <list>
#include <string>

#include <boost/bind.hpp>

namespace DetectDriver
{
const wchar_t* strZwClose = L"ZwClose";
const wchar_t* strNtClose = L"NtClose";

const wchar_t* strZwOpenFile = L"ZwOpenFile";
const wchar_t* strNtOpenFile = L"NtOpenFile";

const wchar_t* strZwQueryDirFile = L"ZwQueryDirectoryFile";
const wchar_t* strNtQueryDirFile = L"NtQueryDirectoryFile";

void GetApiPairPointers(const wchar_t* strZwName,const wchar_t* strNtName,
                        PVOID* sstPtr,PVOID* realPtr)
{
    std::wstring functionNameZw(strZwName);
    UNICODE_STRING functionNameZwUnc = utils::wstring2UnicodeStr(functionNameZw); 

    std::wstring functionNameNt(strNtName);
    UNICODE_STRING functionNameNtUnc = utils::wstring2UnicodeStr(functionNameNt); 

    *sstPtr = utils::GetFunctionSSTPtr(&functionNameZwUnc);
    *realPtr = MmGetSystemRoutineAddress(&functionNameNtUnc);

    if( *realPtr == NULL || *sstPtr == NULL)
        throw std::exception("Can't get system routine address");
}
void GetFileApiPointers(PVOID* sstPtrClose,PVOID* realPtrClose,
                        PVOID* sstPtrOpenFile,PVOID* realPtrOpenFile,
                        PVOID* sstPtrQueryDir,PVOID* realPtrQueryDir)
{
    GetApiPairPointers(strZwClose,strNtClose,sstPtrClose,realPtrClose);
    GetApiPairPointers(strZwOpenFile,strNtOpenFile,sstPtrOpenFile,realPtrOpenFile);
    GetApiPairPointers(strZwQueryDirFile,strNtQueryDirFile,sstPtrQueryDir,realPtrQueryDir);
}

struct FileRecord
{
    std::wstring filePath;
    bool isDir;
};

typedef std::list<FileRecord> FileNameList;

class SaveObserver: public utils::IFileObserver
{
    FileNameList* fileNameList_;
public:
    SaveObserver(FileNameList* fileNameList)
        : fileNameList_(fileNameList)
    {}

    // IFileObserver interfaces
    void OnFileFound(const std::wstring& filePath,DirInfo* info)
    {
        AddNewRecord(filePath,false);
    }
    void OnDirFound(const std::wstring& dirPath,DirInfo* info)
    {
        AddNewRecord(dirPath,true);
    }
private:
    void AddNewRecord(const std::wstring& path,bool isDir)
    {
        FileNameList::iterator it = 
            fileNameList_->insert(fileNameList_->end(),FileRecord());

        it->filePath = path;
        it->isDir = isDir;
    }
};


bool IsFileRecordEqual(const FileRecord& first,const FileRecord& second)
{
    return (first.filePath == second.filePath) && (first.isDir == second.isDir);
}

bool IsListsEqual(const FileNameList& first,const FileNameList& second)
{
    if( first.size() != second.size() )
        return false;

    return std::equal(first.begin(),first.end(),second.begin(),&IsFileRecordEqual);
}

void GetListDifference(const FileNameList& hookedFileList,
                       const FileNameList& originalFileList,
                       FileNameList* hiddenFileList)
{
    FileNameList::const_iterator it = originalFileList.begin();
    for( ; it != originalFileList.end() ; ++it )
    {
        FileNameList::const_iterator res = 
            std::find_if(hookedFileList.begin(),hookedFileList.end(),
                         boost::bind(&IsFileRecordEqual,_1,*it));
        
        if( res != hookedFileList.end() )
            continue;

        hiddenFileList->push_back( *it );
    }
}

class ProxyObserver: public utils::IFileObserver
{
    IDetectObserver* detectObserver_;
public:
    ProxyObserver( IDetectObserver* detectObserver )
        : detectObserver_(detectObserver)
    {}

    // IFileObserver interfaces
    void OnFileFound(const std::wstring& filePath,DirInfo* info)
    {
        detectObserver_->AddHiddenRecord(filePath);
    }
    void OnDirFound(const std::wstring& dirPath,DirInfo* info)
    {
        detectObserver_->AddHiddenRecord(dirPath);        
    }
};

void Detect(const std::wstring& dirPath,
            utils::IEnvFileApi* sstApi,
            utils::IEnvFileApi* realApi,
            IDetectObserver* detectObserver)
{
    FileNameList hookedFileList;
    {
        SaveObserver observer(&hookedFileList);
        utils::FileEnumerator enumerator(&observer,sstApi);
        enumerator.StartEnumeration(dirPath);
    }

    FileNameList originalFileList;
    {
        SaveObserver observer(&originalFileList);
        utils::FileEnumerator enumerator(&observer,realApi);
        enumerator.StartEnumeration(dirPath);
    }

    if( IsListsEqual(hookedFileList,originalFileList) )
        return;

    FileNameList hiddenFileList;
    GetListDifference(hookedFileList,originalFileList,&hiddenFileList);

    // Report result
    FileNameList::const_iterator it = hiddenFileList.begin();
    for( ; it != hiddenFileList.end(); ++it )
    {
        detectObserver->AddHiddenRecord(it->filePath);       

        if( !it->isDir )
            continue;
        
        // Scan all subdirectories and report all result as hidden
        ProxyObserver observer(detectObserver);
        utils::FileSearcher searcher(&observer,realApi);
        searcher.StartSearch(it->filePath);
    }
}

class DirectoryObserver: public utils::IFileObserver
{
    utils::IEnvFileApi* sstApi_;
    utils::IEnvFileApi* realApi_;

    IDetectObserver* detectObserver_;
public:
    DirectoryObserver( utils::IEnvFileApi* sstApi,
                       utils::IEnvFileApi* realApi,
                       IDetectObserver* detectObserver)
        : sstApi_(sstApi)
        , realApi_(realApi)
        , detectObserver_(detectObserver)
    {}

    // IFileObserver interfaces
    void OnFileFound(const std::wstring& filePath,DirInfo* info)
    {
        if( detectObserver_->IsStoped() )
            throw StopException("Detection is stopped.");
    }
    void OnDirFound(const std::wstring& dirPath,DirInfo* info)
    {
        if( detectObserver_->IsStoped() )
            throw StopException("Detection is stopped.");

        detectObserver_->OnDirProcessing(dirPath);

        Detect(dirPath,sstApi_,realApi_,detectObserver_);
    }
};

void DetectHiddenFiles(const std::wstring& path,
                       IDetectObserver* detectObserver)
{
    PVOID sstPtrClose,realPtrClose;
    PVOID sstPtrOpenFile,realPtrOpenFile;
    PVOID sstPtrQueryDir,realPtrQueryDir;

    GetFileApiPointers(&sstPtrClose, &realPtrClose, 
                       &sstPtrOpenFile, &realPtrOpenFile,
                       &sstPtrQueryDir, &realPtrQueryDir);

    if( (sstPtrQueryDir == realPtrQueryDir) && (sstPtrOpenFile==realPtrOpenFile) )
        throw std::exception("No hooks detected.");

    utils::CustomFileApi sstFileApi(sstPtrClose,sstPtrOpenFile,sstPtrQueryDir);
    utils::CustomFileApi realFileApi(realPtrClose,realPtrOpenFile,realPtrQueryDir);

    // Scan root directory
    detectObserver->OnDirProcessing(path);
    Detect(path,&sstFileApi,&realFileApi,detectObserver);

    // Scan all subdirectories
    DirectoryObserver dirObserver(&sstFileApi,&realFileApi,detectObserver);
    utils::FileSearcher searcher(&dirObserver,&realFileApi);
    searcher.StartSearch(path);
}

}