#include "drvCommon.h"
#include "DetectDriver_Ioctl.h"

#include "FileDetector.h"
#include "FileDetectorImpl.h"
#include "PackageParser.h"
#include "KernelHandleGuard.h"
#include "KernelNames.h"

#include <string>
#include <stdexcept>

namespace
{
    DetectDriver::FileDetectorImpl fileDetectorImpl;
}

namespace DetectDriver
{

// Information presentation
void PackInfo( const WStringList& hiddenList, 
			   WCHAR* buf, ULONG bufSize, ULONG *bytesTxd)
{
	if(hiddenList.empty())
		throw std::exception("Empty hidden file list");

	std::wstring procListString;
	utils::PackByEOL(hiddenList,&procListString);

	if(procListString.empty())
		throw std::exception("Internal error occur while packing params.");

	if(procListString.size() > bufSize)
		throw std::exception("Buffer size is not enough to hold hidden process list.");

	memcpy(buf,procListString.c_str(),procListString.size()*2);
	*bytesTxd = procListString.size()*2;
}

// User communication

void FileDetector::Initialize(utils::QueryMng& queryMng)
{
    queryMng.Subscribe(IOCTL_FILES_START_DETECT,this);
    queryMng.Subscribe(IOCTL_FILES_STOP_DETECT,this);
    queryMng.Subscribe(IOCTL_FILES_QUERY_HIDEN,this);
    queryMng.Subscribe(IOCTL_FILES_QUERY_STATUS,this);
    queryMng.Subscribe(IOCTL_FILES_QUERY_ERROR,this);
}

void FileDetector::Cleanup(utils::QueryMng& queryMng)
{
    queryMng.UnSubscribe(IOCTL_FILES_START_DETECT,this);
    queryMng.UnSubscribe(IOCTL_FILES_STOP_DETECT,this);
    queryMng.UnSubscribe(IOCTL_FILES_QUERY_HIDEN,this);
    queryMng.UnSubscribe(IOCTL_FILES_QUERY_STATUS,this);
    queryMng.UnSubscribe(IOCTL_FILES_QUERY_ERROR,this);
}

void DispatchQueryStatus(WCHAR* buf,
                         ULONG outputBufSize,
                         ULONG* bytesTxd)
{
    if( outputBufSize < 3*sizeof(ULONG) )
        throw std::exception("Buffer size is not enough to hold status information.");

    ULONG status,flag,totalCount;
    fileDetectorImpl.QueryStatus(&status,&flag,&totalCount);
    memcpy((char*)buf                ,&status    ,sizeof(ULONG));
    memcpy((char*)buf+1*sizeof(ULONG),&flag      ,sizeof(ULONG));
    memcpy((char*)buf+2*sizeof(ULONG),&totalCount,sizeof(ULONG));
    *bytesTxd = 3*sizeof(ULONG);
}

void DispatchQueryHidden(WCHAR* buf,
                         ULONG outputBufSize,
                         ULONG* bytesTxd)
{
    FilePathList filePathList;
    fileDetectorImpl.QueryHiddenFiles(&filePathList);
    PackInfo(filePathList,buf,outputBufSize,bytesTxd);
}



void DispatchImpl(ULONG controlCode,
                  WCHAR* inputBuf,
                  ULONG inputBufSize,
                  WCHAR* outputBuf,
                  ULONG outputBufSize,
                  ULONG* bytesTxd)
{
    switch(controlCode)
    {
    case IOCTL_FILES_START_DETECT:
        {
            wchar_t* bufEnd = (wchar_t*)((char*)inputBuf+inputBufSize);
            std::wstring userPath(inputBuf,bufEnd);
            std::wstring kernelPath = utils::ConvertUserToKernelPath(userPath);
            fileDetectorImpl.Start(kernelPath);
        }
        break;
    case IOCTL_FILES_STOP_DETECT:
        fileDetectorImpl.Stop();
        break;
    case IOCTL_FILES_QUERY_STATUS:
        DispatchQueryStatus(outputBuf,outputBufSize,bytesTxd);
        break;
    case IOCTL_FILES_QUERY_HIDEN:
        DispatchQueryHidden(outputBuf,outputBufSize,bytesTxd);
        break;
    case IOCTL_FILES_QUERY_ERROR:
        {
            std::string error;
            fileDetectorImpl.QueryErrorString(&error);
            throw std::runtime_error(error);
        }
        break;
    }
}

void FileDetector::Dispatch(ULONG controlCode,
							WCHAR* buf,
							ULONG inputBufSize,
							ULONG outputBufSize,
							ULONG* bytesTxd)
{
	DbgPrint("-DetectDriver- Query Hidden Files\n");

	// Need to dispatch exception message by hand
	// to provide information type that will be returned

	// First byte for information type
	wchar_t* pOutputBuf = (wchar_t*)( (char*)buf + 1 );
	try
	{
		DispatchImpl(controlCode,buf,inputBufSize,pOutputBuf,outputBufSize-1,bytesTxd);
	}
	catch(const std::exception& ex)
	{      
        const char* errorStr = ex.what(); // For debug
        size_t strLen = strlen(errorStr);
        if( strLen +1 > outputBufSize)
            throw; // Exception will be dispatched by QueryManager

        // Signal that call is fail
        char infoType = QUERY_FAIL;
        memcpy(buf,&infoType,1);
        *bytesTxd = 1;

        // Copy exception string
        memcpy(pOutputBuf,errorStr,strLen);
        *bytesTxd += strLen;
        return;
	}

    // Signal that call is successful
    char infoType = QUERY_SUCCESS;
    memcpy(buf,&infoType,1);
    *bytesTxd += 1;
}

}//namespace DetectDriver