#include "drvCommon.h"
#include "DetectDriver_Ioctl.h"

#include "FileDetectorImpl.h"
#include "FileDetectAlgorithm.h"

namespace DetectDriver
{

class DetectObserver:public IDetectObserver
{
    FileDetectorImpl* owner_;
public:
    DetectObserver(FileDetectorImpl* owner)
        : owner_(owner)
    {}
      
    void AddHiddenRecord(const std::wstring& path)
    {
        owner_->AddHiddenRecord(path);
    }

    void OnDirProcessing(const std::wstring& dirPath)
    {
        owner_->OnDirProcessing(dirPath);
    }
   
    bool IsStoped()const
    {
        return owner_->IsStoped();
    }
};

void FileDetectorImpl::AddHiddenRecord(const std::wstring& path)
{
    utils::AutoMutex guard(&lock_);
    hiddenFilesList_.push_back(path);
    flag_ = FILES_DETECT_FLAG_NEW_HIDDEN_FILE;
}

void FileDetectorImpl::OnDirProcessing(const std::wstring& dirPath)
{
    InterlockedIncrement(&totalCount_);
}

bool FileDetectorImpl::IsStoped()const
{
    return !isRunning_;
}

static void ThreadWorkerRoutine( void* context )
{
    FileDetectorImpl* pImpl = (FileDetectorImpl*)context;
    pImpl->ThreadRoutine();
}

void FileDetectorImpl::ThreadRoutine()
{
    try
    {
        std::wstring startPath;
        {
            utils::AutoMutex guard(&lock_);
            startPath = startPath_;
        }

        if( startPath.empty() )
            throw std::exception("Path can't be empty.");

        std::wstring::iterator it = startPath.end();
        if( *(--it) != L'\\' )
            startPath += L'\\';

        DetectObserver observer(this);
        DetectHiddenFiles(startPath.c_str(),&observer);
    }
    catch(const StopException&)
    {} // Just skip
    catch(const std::exception& ex)
    {
        KdPrint(( __FUNCTION__" fail. Exception: %s",ex.what() ));

        utils::AutoMutex guard(&lock_);
        isRunning_= false;
        status_ = FILES_DETECT_STATUS_ABORTED;
        flag_ = FILES_DETECT_FLAG_ERROR;
        errorString_ = ex.what(); 
        return;
    }

    // Update status
    {
        utils::AutoMutex guard(&lock_);
        isRunning_= false;
        status_ = FILES_DETECT_STATUS_FINISHED;
    }
}

FileDetectorImpl::FileDetectorImpl()
    : isRunning_(FALSE)
    , flag_(FILES_DETECT_FLAG_EMPTY)
    , status_(FILES_DETECT_STATUS_EMPTY)
    , totalCount_(0)
{}

void FileDetectorImpl::Start(const std::wstring& path)
{
    if( InterlockedCompareExchange(&isRunning_,TRUE,FALSE)== TRUE)
        throw std::exception("Detection already started.");

    // Minimal lock
    {
        utils::AutoMutex guard(&lock_);
        status_ = FILES_DETECT_STATUS_RUNNING;
        startPath_ = path;
        hiddenFilesList_.clear();
    }

    HANDLE hThread = NULL;
    NTSTATUS status = 
        PsCreateSystemThread(&hThread,
                             THREAD_ALL_ACCESS,
                             NULL,
                             NULL,
                             NULL,
                             &ThreadWorkerRoutine,
                             this);
    if( !NT_SUCCESS(status) )
    {
        InterlockedExchange(&isRunning_,FALSE);
        utils::AutoMutex guard(&lock_);
        status_ = FILES_DETECT_STATUS_FINISHED;
        throw std::exception("PsCreateSystemThread fail.\n");
    }
    ZwClose(hThread);
}

void FileDetectorImpl::Stop()
{
    if( InterlockedCompareExchange(&isRunning_,FALSE,TRUE)==FALSE)
        throw std::exception("Detection not started.");

    utils::AutoMutex guard(&lock_);
    status_ = FILES_DETECT_STATUS_ABORTED;
}

void FileDetectorImpl::QueryStatus(ULONG* status, ULONG* flag, ULONG* totalCount)
{
    utils::AutoMutex guard(&lock_);

    *status = status_;
    *flag = flag_;
    *totalCount = totalCount_;
}

void FileDetectorImpl::QueryHiddenFiles(FilePathList* filePathList)
{
    utils::AutoMutex guard(&lock_);

    if( hiddenFilesList_.empty() )
        return;

    filePathList->swap(hiddenFilesList_);
    flag_ = FILES_DETECT_FLAG_EMPTY;
}

void FileDetectorImpl::QueryErrorString(std::string* error)
{
    utils::AutoMutex guard(&lock_);

    *error = errorString_;
}

}